//Free library of highly optimized functions written
//specially for KCompiler component

{Copyright  2002-2003 by DOMIN Software}

{-----------------------------------------------------------------------
References:

1) Borland Delphi Runtime Library, Math Unit
2) "IA-32 Intel Architecture Software Developer's Manual, Volume 2:
   Instruction Set Reference", Intel, 2002
3) "Intel Pentium 4 and Intel(R) Xeon Processor Optimization
   Reference Manual", Intel, 2001
4) "Intel Processor Identification and the CPUID Instruction", Intel, 2002

All angle parameters and results of trig functions are in radians.

Most of the following routines map directly to Intel 80387 FPU floating point
machine instructions.  Input domains, output ranges, and error handling are
determined largely by the FPU hardware. Routines favor the Pentium III FPU
pipeline architecture.
-----------------------------------------------------------------------}
unit KLib;

interface
uses
  KCompiler;

var
{ This variable indicates whether the Pentium Pro instructions are supported }
{ and will be set automatically during unit initialization. Change it manually }
{ if you wish to control behavior of "min" and "max" functions }
  IsPPro : boolean = false;

{ Trigonometric functions }

function RegisterSin(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterCos(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterTan(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterCotan(AKCompiler : TKCompiler; alias : string) : boolean;

function RegisterArcSin(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterArcCos(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterArcTan(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterArcCotan(AKCompiler : TKCompiler; alias : string) : boolean;

{ Hyperbolic functions and inverses }

function RegisterCosh(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterSinh(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterTanh(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterCotanh(AKCompiler : TKCompiler; alias : string) : boolean;

function RegisterArcCosh(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterArcSinh(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterArcTanh(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterArcCotanh(AKCompiler : TKCompiler; alias : string) : boolean;

{ Selector functions }

function RegisterMin(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterMax(AKCompiler : TKCompiler; alias : string) : boolean;

{ Miscellaneous functions }

function RegisterIntPower(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterPower(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterAbs(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterSqrt(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterSqr(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterExp(AKCompiler : TKCompiler; alias : string) : boolean;

{ Approximating function }

function RegisterInt(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterFrac(AKCompiler : TKCompiler; alias : string) : boolean;
function RegisterRound(AKCompiler : TKCompiler; alias : string) : boolean;

{ Logarithmic functions }

{ Log base e of X }
function RegisterLn(AKCompiler : TKCompiler; alias : string) : boolean;
{ Log base 2 of X }
function RegisterLog2(AKCompiler : TKCompiler; alias : string) : boolean;
{ Log base 10 of X }
function RegisterLog10(AKCompiler : TKCompiler; alias : string) : boolean;
{ Log base N of X }
function RegisterLogN(AKCompiler : TKCompiler; alias : string) : boolean;

{ Constants }

{ 1 }
function RegisterOne(AKCompiler : TKCompiler; alias : string) : boolean;
{ Log base 2 of 10 }
function RegisterL2T(AKCompiler : TKCompiler; alias : string) : boolean;
{ Log base 2 of e }
function RegisterL2E(AKCompiler : TKCompiler; alias : string) : boolean;
{ Pi }
function RegisterPi(AKCompiler : TKCompiler; alias : string) : boolean;
{ Log base 10 of 2 }
function RegisterLg2(AKCompiler : TKCompiler; alias : string) : boolean;
{ Log base e of 2 }
function RegisterLn2(AKCompiler : TKCompiler; alias : string) : boolean;
{ 0 }
function RegisterZero(AKCompiler : TKCompiler; alias : string) : boolean;

implementation
var
  cwChop : Word = $1F32;
  half : Single = 0.5;
  MaxExp : Double = MaxInt;

procedure _int{(x : extended) : extended};
asm
  fstcw [esp-4]
  fldcw cwChop
  frndint
  fldcw [esp-4]
end;

procedure _frac{(x : extended) : extended};
asm
  fld st(0)
  fstcw [esp-4]
  fldcw cwChop
  frndint
  fldcw [esp-4]
  fsubp
end;

procedure _exp{(x : extended) : extended};
asm
  fldl2e
  fmul
  fld st(0)
  frndint
  fsub st(1),st(0)
  fxch
  f2xm1
  fld1
  fadd
  fscale
  fstp st(1)
end;

procedure _Cosh{(x : extended) : extended};
asm
  call _exp
  fld st(0)
  fld1
  fdivr st(1),st(0)
  fadd st(0),st(0)
  fxch st(2)
  faddp st(1),st(0)
  fdivrp st(1),st(0)
end;

procedure _Sinh{(x : extended) : extended};
asm
  call _exp
  fld st(0)
  fld1
  fdivr st(1),st(0)
  fadd st(0),st(0)
  fxch st(2)
  fsubrp st(1),st(0)
  fdivrp st(1),st(0)
end;

procedure _Tanh{(x : extended) : extended};
asm
  fadd st(0),st(0)
  call _exp
  fld st(0)
  fld1
  fadd st(1),st(0)
  fsubp st(2),st(0)
  fdivp
end;

procedure _Cotanh{(x : extended) : extended};
asm
  fadd st(0),st(0)
  call _exp
  fld st(0)
  fld1
  fadd st(1),st(0)
  fsubp st(2),st(0)
  fdivrp
end;

procedure _ArcCosh{(x : extended) : extended};
asm
  fld st(0)
  fmul st(0), st(0)
  fld1
  fsubp
  fsqrt
  faddp
  fldln2
  fxch
  fyl2x
end;

procedure _ArcSinh{(x : extended) : extended};
asm
  xor edx, edx
  ftst
  fstsw ax
  sahf
  jnb @@NotNeg
  mov edx, 1
@@NotNeg:
  fabs
  fld st(0)
  fmul st(0), st(0)
  fld1
  faddp
  fsqrt
  faddp
  fldln2
  fxch
  fyl2x
  test edx, edx
  jz @@exit
  fchs
@@exit:
end;

procedure _ArcTanh{(x : extended) : extended};
asm
  xor edx, edx
  ftst
  fstsw ax
  sahf
  jnb @@NotNeg
  mov edx, 1
@@NotNeg:
  fabs
  fld st(0)
  fld1
  fadd st(2), st(0)
  fsubrp st(1), st(0)
  fdivp
  fldln2
  fxch
  fyl2x
  fmul half
  test edx, edx
  jz @@exit
  fchs
@@exit:
end;

procedure _ArcCotanh{(x : extended) : extended};
asm
  fld1
  fdivrp
  call _ArcTanh
end;

procedure _ArcCos{(x : extended) : extended};
asm
  fld st(0)
  fmul st(0),st(0)
  fld1
  fsubrp
  fsqrt
  fxch st(1)
  fpatan
end;

procedure _ArcSin{(x : extended) : extended};
asm
  fld st(0)
  fmul st(0),st(0)
  fld1
  fsubrp
  fsqrt
  fpatan
end;

procedure _Min{(a = st(1),b = st(0): extended) : extended};
asm
  fcom
  fstsw ax
  sahf
  jb @@1
  fstp st(1)
  jmp @@exit
@@1:
  fstp st(0)
@@exit:
end;

procedure _Max{(a = st(1),b = st(0): extended) : extended};
asm
  fcom
  fstsw ax
  sahf
  jnbe @@1
  fstp st(1)
  jmp @@exit
@@1:
  fstp st(0)
@@exit:
end;

procedure _IntPower{(Base: Extended; Exponent: Integer): Extended};
asm
  mov ecx, eax
  cdq
  fld1
  fxch st(1)
  xor eax, edx
  sub eax, edx
  jz @@4
  jmp @@2
@@1:
  fmul st(0), st(0)
@@2:
  shr eax,1
  jnc @@1
  fmul st(1), st(0)
  jnz @@1
  fstp st
  cmp ecx, 0
  jge @@3
  fld1
  fdivrp
  jmp @@3
@@4:
  fstp st(0)
@@3:
end;

procedure _Power{(Base = st(1), Exponent = st(0): Extended): Extended};
asm
//if (Exponent > 0.0) and (Base = 0.0)
  ftst
  fstsw ax
  sahf
  jna @@NegExp
  fxch st(1)
  ftst
  fstsw ax
  sahf
  jne @@NotZero
  fstp st(1)
  jmp @@exit
@@NotZero:
  fxch st(1)
@@NegExp:
  fld st(0)
  frndint
  fcomp
  fstsw ax
  sahf
  jne @@Real
  fld st(0)
  fabs
  fcomp MaxExp
  ja @@Real
  fistp dword ptr [esp-4]
  mov eax, [esp-4]
  call _IntPower
  jmp @@exit
@@Real:
  fxch st(1)
  fldln2
  fxch
  fyl2x
  fmulp
  call _exp
@@exit:
end;

procedure _LogN{(Base = st(1), X = st(0): Extended): Extended};
asm
  fld1
  fxch
  fyl2x
  fxch
  fld1
  fxch
  fyl2x
  fdivp
end;

function RegisterSin(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    0, nil, true, [$D9, $FE]) <> 0 //fsin
  else Result := false;
end;

function RegisterCos(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    0, nil, true, [$D9, $FF]) <> 0 //fcos
  else Result := false;
end;

function RegisterTan(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    1, nil, true, [$D9, $F2, $DD, $D8]) <> 0 //fptan; fstp st(0)
  else Result := false;
end;

function RegisterCotan(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    1, nil, true, [$D9, $FB, $DE, $F1]) <> 0 // fsincos; fdivrp
  else Result := false;
end;

function RegisterArcSin(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    2, @_ArcSin, true, []) <> 0
  else Result := false;
end;

function RegisterArcCos(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    2, @_ArcCos, true, []) <> 0
  else Result := false;
end;

function RegisterArcTan(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    1, nil, true, [$D9, $E8, $D9, $F3]) <> 0 //fld1; fpatan
  else Result := false;
end;

function RegisterArcCotan(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    1, nil, true, [$D9, $E8, $D9, $C9, $D9, $F3]) <> 0 //fld1; fxch; fpatan
  else Result := false;
end;

function RegisterCosh(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    2, @_Cosh, true, []) <> 0
  else Result := false;
end;

function RegisterSinh(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    2, @_Sinh, true, []) <> 0
  else Result := false;
end;

function RegisterTanh(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    2, @_ArcTanh, true, []) <> 0
  else Result := false;
end;

function RegisterCotanh(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    2, @_ArcCotanh, true, []) <> 0
  else Result := false;
end;

function RegisterArcCosh(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    2, @_ArcCosh, true, []) <> 0
  else Result := false;
end;

function RegisterArcSinh(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    2, @_Sinh, true, []) <> 0
  else Result := false;
end;

function RegisterArcTanh(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    2, @_Tanh, true, []) <> 0
  else Result := false;
end;

function RegisterArcCotanh(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    2, @_CoTanh, true, []) <> 0
  else Result := false;
end;

function RegisterMin(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    if IsPPro then
      Result := AKCompiler.RegisterFuncExt(alias, 'vf80vf80', ctBuilt_in, frExtended,
      0, nil, true, [$DB, $F1, $DB, $D1, $DD, $D9]) <> 0 //fcomi; fcmovnbe; fstp st(1)
    else
      Result := AKCompiler.RegisterFuncExt(alias, 'vf80vf80', ctBuilt_in, frExtended,
      0, @_min, true, []) <> 0
  else Result := false;
end;

function RegisterMax(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    if IsPPro then
      Result := AKCompiler.RegisterFuncExt(alias, 'vf80vf80', ctBuilt_in, frExtended,
      0, nil, true, [$DB, $F1, $DA, $C1, $DD, $D9]) <> 0 //fcomi; fcmovb; fstp st(1)
    else
      Result := AKCompiler.RegisterFuncExt(alias, 'vf80vf80', ctBuilt_in, frExtended,
      0, @_max, true, []) <> 0
  else Result := false;
end;

function RegisterIntPower(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80vi32', ctBuilt_in, frExtended,
    1, @_IntPower, true, []) <> 0
  else Result := false;
end;

function RegisterPower(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80vf80', ctBuilt_in, frExtended,
    2, @_Power, true, []) <> 0
  else Result := false;
end;

function RegisterAbs(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    0, nil, true, [$D9, $E1]) <> 0 //fabs
  else Result := false;
end;

function RegisterSqrt(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    0, nil, true, [$D9, $FA]) <> 0 //fsqrt
  else Result := false;
end;

function RegisterSqr(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    0, nil, true, [$D8, $C8]) <> 0 //fmul st(0), st(0)
  else Result := false;
end;

function RegisterExp(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    2, @_exp, true, []) <> 0
  else Result := false;
end;

function RegisterInt(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    0, @_int, true, []) <> 0
  else Result := false;
end;

function RegisterFrac(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    1, @_frac, true, []) <> 0
  else Result := false;
end;

function RegisterRound(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    0, nil, true, [$D9, $FC]) <> 0 //frndint
  else Result := false;
end;

function RegisterLn(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    1, nil, true, [$D9, $ED, $D9, $C9, $D9, $F1]) <> 0 //fldln2; fxch; fyl2x
  else Result := false;
end;

function RegisterLog2(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    1, nil, true, [$D9, $E8, $D9, $C9, $D9, $F1]) <> 0 //fld1; fxch; fyl2x
  else Result := false;
end;

function RegisterLog10(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80', ctBuilt_in, frExtended,
    1, nil, true, [$D9, $EC, $D9, $C9, $D9, $F1]) <> 0 //fldlg2; fxch; fyl2x
  else Result := false;
end;

function RegisterLogN(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, 'vf80vf80', ctBuilt_in, frExtended,
    1, @_logN, true, []) <> 0
  else Result := false;
end;

function RegisterOne(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, '', ctBuilt_in, frExtended,
    0, nil, false, [$D9, $E8]) <> 0 //fld1
  else Result := false;
end;

function RegisterL2T(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, '', ctBuilt_in, frExtended,
    0, nil, false, [$D9, $E9]) <> 0 //fldl2t
  else Result := false;
end;

function RegisterL2E(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, '', ctBuilt_in, frExtended,
    0, nil, false, [$D9, $EA]) <> 0 //fldl2e
  else Result := false;
end;

function RegisterPi(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, '', ctBuilt_in, frExtended,
    0, nil, false, [$D9, $EB]) <> 0 //fldpi
  else Result := false;
end;

function RegisterLg2(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, '', ctBuilt_in, frExtended,
    0, nil, false, [$D9, $EC]) <> 0 //fldlg2
  else Result := false;
end;

function RegisterLn2(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, '', ctBuilt_in, frExtended,
    0, nil, false, [$D9, $ED]) <> 0 //fldln2
  else Result := false;
end;

function RegisterZero(AKCompiler : TKCompiler; alias : string) : boolean;
begin
  if Assigned(AKCompiler) then
    Result := AKCompiler.RegisterFuncExt(alias, '', ctBuilt_in, frExtended,
    0, nil, false, [$D9, $EE]) <> 0 //fldz
  else Result := false;
end;


begin
  asm
//test whether we can use cpuid instruction
    push ebx
    pushfd
    pop eax
    mov ecx,eax
    xor eax,0200000h
    push eax
    popfd
    pushfd
    pop eax
    xor eax,ecx
    jz @@exit
//test whether we can use Pentium Pro instructions
    mov eax,1
    dw 0a20fh //cpuid
    test edx, 08001h
    jz @@exit
    mov IsPPro, true
@@exit:
    pop ebx
  end;
end.
